# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

__version__: str
version: str
__version_tuple__: tuple[int, int, int, str, str] | tuple[int, int, int]
version_tuple: tuple[int, int, int, str, str] | tuple[int, int, int]
